@@ -1,6 +1,6 @@
 package Types::Numbers;
 
-our $VERSION = '0.91'; # VERSION
+our $VERSION = '0.92'; # VERSION
 # ABSTRACT: Type constraints for numbers
 
 #############################################################################
@@ -19,16 +19,13 @@ use Types::Standard ();
 
 use Scalar::Util v1.20 (qw(blessed looks_like_number));  # support for overloaded/blessed looks_like_number
 use POSIX 'ceil';
+use Math::BigInt;
 use Math::BigFloat;
 use Data::Float;
 use Data::Integer;
 
 use constant {
-   _BASE2_LOG     => log(2) / log(10),
-   _SAFE_NUM_MIN  => Data::Integer::min_signed_natint   < Data::Float::max_integer * -1 ?
-                     Data::Integer::min_signed_natint   : Data::Float::max_integer * -1,
-   _SAFE_NUM_MAX  => Data::Integer::max_unsigned_natint > Data::Float::max_integer *  1 ?
-                     Data::Integer::max_unsigned_natint : Data::Float::max_integer *  1,
+   _BASE2_LOG => log(2) / log(10),
 };
 
 sub _croak ($;@) { require Type::Exception; goto \&Type::Exception::croak }
@@ -41,11 +38,28 @@ no warnings;  # don't warn on type checks
 my $bigtwo = Math::BigFloat->new(2);
 my $bigten = Math::BigFloat->new(10);
 
+# Large 64-bit floats (long doubles) tend to stringify themselves in exponent notation, even
+# though the number is still pristine.  IOW, the numeric form is perfect, but the string form
+# loses information.  This can be a problem for stringified inlines.
+my @df_max_int_parts = Data::Float::float_parts( Data::Float::max_integer );
+my $DF_MAX_INT = $bigtwo->copy->bpow($df_max_int_parts[1])->bmul($df_max_int_parts[2])->as_int;
+
+my $SAFE_NUM_MIN = Math::BigInt->new(
+   Data::Integer::min_signed_natint   < $DF_MAX_INT * -1 ?
+   Data::Integer::min_signed_natint   : $DF_MAX_INT * -1
+);
+my $SAFE_NUM_MAX = Math::BigInt->new(
+   Data::Integer::max_unsigned_natint > $DF_MAX_INT *  1 ?
+   Data::Integer::max_unsigned_natint : $DF_MAX_INT *  1,
+);
+
 my $meta = __PACKAGE__->meta;
 
 #############################################################################
 # Framework types
 
+### TODO: Coercions where safe ###
+
 # Moose and Type::Tiny types both don't seem to support Math::Big* = Num.
 # So, we have to start almost from stratch.
 my $_NumLike = $meta->add_type(
@@ -92,19 +106,20 @@ my $_NumRange_perlsafe = Type::Tiny->new(
    display_name => "_NumRange_perlsafe",
    parent     => $_NumLike,
    # no equals because MAX+1 = MAX after truncation
-   constraint => sub { $_ > _SAFE_NUM_MIN and $_ < _SAFE_NUM_MAX },
-   inlined    => sub { "$_[1] > "._SAFE_NUM_MIN." and $_[1] < "._SAFE_NUM_MAX },
+   constraint => sub { $_ > $SAFE_NUM_MIN and $_ < $SAFE_NUM_MAX },
+   inlined    => sub { "$_[1] > ".$SAFE_NUM_MIN." and $_[1] < ".$SAFE_NUM_MAX },
 );
 
-### XXX: This string equality check is necessary because Math::BigInt seems to think 1.5 == 1
+### XXX: This string equality check is necessary because Math::BigInt seems to think 1.5 == 1.
+### However, this is problematic with long doubles that stringify into E notation.
 my $_IntLike = $meta->add_type(
    name       => 'IntLike',
    parent     => $_NumLike,
    library    => __PACKAGE__,
-   constraint => sub { /\d+/ and int($_) == $_ and int($_) eq $_ },
+   constraint => sub { /\d+/ and int($_) == $_ and (int($_) eq $_ or not ref($_)) },
    inlined    => sub {
       my ($self, $val) = @_;
-      $self->parent->inline_check($val)." and $val =~ /\\d+/ and int($val) == $val and int($val) eq $val";
+      $self->parent->inline_check($val)." and $val =~ /\\d+/ and int($val) == $val and (int($val) eq $val or not ref($val))";
    },
 );
 
@@ -273,7 +288,10 @@ my $_PerlSafeInt = $meta->add_type( Type::Tiny::Intersection->new(
 )->create_child_type(
    name       => 'PerlSafeInt',
    library    => __PACKAGE__,
-   inlined    => sub { "!ref($_[1]) and ".$_IntLike->inline_check($_[1]).' and '.$_NumRange_perlsafe->inline_check($_[1]) },
+   inlined    => sub {
+      my $val = $_[1];
+      "!ref($val) and $val =~ /\\d+/ and int($val) == $val and ".$_NumRange_perlsafe->inline_check($val);
+   },
 ) );
 
 my $_BlessedInt = $meta->add_type( Type::Tiny::Intersection->new(
@@ -282,7 +300,10 @@ my $_BlessedInt = $meta->add_type( Type::Tiny::Intersection->new(
 )->create_child_type(
    name       => 'BlessedInt',
    library    => __PACKAGE__,
-   inlined    => sub { Types::Standard::Object->inline_check($_[1]).' and '.$_IntLike->inline_check($_[1]) },
+   inlined    => sub {
+      my $val = $_[1];
+      Types::Standard::Object->inline_check($val)." and $val =~ /\\d+/ and int($val) == $val and int($val) eq $val";
+   },
    constraint_generator => sub {
       my $self = $Type::Tiny::parameterize_type;
       my $digits = shift;
@@ -369,7 +390,7 @@ $meta->add_type(
       my $_NumRange_param   = $_NumRange  ->parameterize($min, $max);
 
       # inline will already have the IntLike check, and maybe not need the extra NumRange check
-      my $perlsafe_inline = $min >= _SAFE_NUM_MIN && $max <= _SAFE_NUM_MAX ?
+      my $perlsafe_inline = $min >= $SAFE_NUM_MIN && $max <= $SAFE_NUM_MAX ?
          sub { Types::Standard::Str->inline_check($_[0]) } :
          sub { '('.Types::Standard::Str->inline_check($_[0]).' and '.$_NumRange_perlsafe->inline_check($_[0]).')' }
       ;
@@ -760,217 +781,191 @@ The hierarchy of the types is as follows:
 
 =head2 Basic types
 
-=over
+=head3 NumLike
 
-=item C<< NumLike >>
+Behaves like C<<< LaxNum >>> from L<Types::Standard>, but will also accept blessed number types.  Unlike C<<< StrictNum >>>, it will accept C<<< NaN >>>
+and C<<< Inf >>> numbers.
 
-Behaves like C<LaxNum> from L<Types::Standard>, but will also accept blessed number types.  Unlike C<StrictNum>, it will accept C<NaN>
-and C<Inf> numbers.
-
-=item C<< NumRange[`n, `p] >>
+=head3 NumRange[`n, `p]
 
 Only accepts numbers within a certain range.  The two parameters are the minimums and maximums, inclusive.
 
-=item C<< PerlNum >>
+=head3 PerlNum
 
-Exactly like C<LaxNum>, but with a different parent.  Only accepts unblessed numbers.
+Exactly like C<<< LaxNum >>>, but with a different parent.  Only accepts unblessed numbers.
 
-=item C<< BlessedNum >>
+=head3 BlessedNum
 
 Only accepts blessed numbers.  A blessed number would be using something like L<Math::BigInt> or L<Math::BigFloat>.  It doesn't
-directly C<isa> check those classes, just that the number is blessed.
+directly C<<< isa >>> check those classes, just that the number is blessed.
 
-=item C<< BlessedNum[`d] >>
+=head3 BlessedNum[`d]
 
-A blessed number that supports at least certain amount of digit accuracy.  The blessed number must support the C<accuracy> or
-C<div_scale> method.
+A blessed number that supports at least certain amount of digit accuracy.  The blessed number must support the C<<< accuracy >>> or
+C<<< div_scale >>> method.
 
-For example, C<< BlessedNum[40] >> would work for the default settings of L<Math::BigInt>, and supports numbers at least as big as
+For example, C<<< BlessedNum[40] >>> would work for the default settings of L<Math::BigInt>, and supports numbers at least as big as
 128-bit integers.
 
-=item C<< NaN >>
+=head3 NaN
 
-A "not-a-number" value, either embedded into the Perl native float or a blessed C<NaN>, checked via C<is_nan>.
+A "not-a-number" value, either embedded into the Perl native float or a blessed C<<< NaN >>>, checked via C<<< is_nan >>>.
 
-=item C<< Inf >>
+=head3 Inf
 
-An infinity value, either embedded into the Perl native float or a blessed C<Inf>, checked via C<is_inf>.
+An infinity value, either embedded into the Perl native float or a blessed C<<< Inf >>>, checked via C<<< is_inf >>>.
 
-=item C<< Inf[`s] >>
+=head3 Inf[`s]
 
-   Inf['+']
-   Inf['-']
+    Inf['+']
+    Inf['-']
 
-An infinity value with a certain sign, either embedded into the Perl native float or a blessed C<Inf>, checked via C<is_inf>.  The
+An infinity value with a certain sign, either embedded into the Perl native float or a blessed C<<< Inf >>>, checked via C<<< is_inf >>>.  The
 parameter must be a plus or minus character.
 
-=item C<< RealNum >>
+=head3 RealNum
 
-Like C<NumLike>, but does not accept NaN or Inf.  Closer to the spirit of C<StrictNum>, but accepts blessed numbers as well.
-
-=back
+Like L</NumLike>, but does not accept NaN or Inf.  Closer to the spirit of C<<< StrictNum >>>, but accepts blessed numbers as well.
 
 =head2 Integers
 
-=over
+=head3 IntLike
 
-=item C<< IntLike >>
+Behaves like C<<< Int >>> from L<Types::Standard>, but will also accept blessed number types and integers in E notation.  There are no
+expectations of storage limitations here.  (See L</SignedInt> for that.)
 
-Behaves like C<Int> from L<Types::Standard>, but will also accept blessed number types and integers in E notation.  There are no
-expectations of storage limitations here.  (See C<SignedInt> for that.)
-
-=item C<< PerlSafeInt >>
+=head3 PerlSafeInt
 
 A Perl (unblessed) integer number than can safely hold the integer presented.  This varies between 32-bit and 64-bit versions of Perl.
 
 For example, for most 32-bit versions of Perl, the largest integer than can be safely held in a 4-byte NV (floating point number) is
-C<18446744073709551614>.  Numbers can go higher than that, but due to the NV's mantissa length (accuracy), information is lost beyond
+C<<< 9007199254740992 >>>.  Numbers can go higher than that, but due to the NV's mantissa length (accuracy), information is lost beyond
 this point.
 
-In this case, C<...614> would pass and C<...615> would fail.
+In this case, C<<< ...992 >>> would pass and C<<< ...993 >>> would fail.
 
-(Technically, the max integer is C<...615>, but we can't tell the difference between C<...615> and C<...616>, so the cut
-off point is C<...614>, inclusive.)
+(Technically, the max integer is C<<< ...993 >>>, but we can't tell the difference between C<<< ...993 >>> and C<<< ...994 >>>, so the cut
+off point is C<<< ...992 >>>, inclusive.)
 
-=item C<< BlessedInt >>
+Be aware that Perls compiled with "long doubles" have a unique problem with storage and information loss: their number form maintains
+accuracy while their stringified form loses information.  For example, take the max safe integer for a long double:
+
+    $num = 18446744073709551615;
+    say $num;  # 1.84467440737095516e+19
+    say $num == 18446744073709551615;  # true, so the full number is still there
+
+These numbers are considered safe for storage.  If this is not preferred, consider a simple C<<< /e/ >>> check for stringified E notation.
+
+=head3 BlessedInt
 
 A blessed number than is holding an integer.  (A L<Math::BigFloat> with an integer value would still pass.)
 
-=item C<< BlessedInt[`d] >>
+=head3 BlessedInt[`d]
 
-A blessed number holding an integer of at most C<`d> digits (inclusive).  The blessed number container must also have digit accuracy
-to support this number.  (See C<< BlessedNum[`d] >>.)
+A blessed number holding an integer of at most C<<< `d >>> digits (inclusive).  The blessed number container must also have digit accuracy
+to support this number.  (See L</BlessedNum[`d]>.)
 
-=item C<< SignedInt >>
+=head3 SignedInt
 
-A signed integer (blessed or otherwise) that can safely hold its own number.  This is different than C<IntLike>, which doesn't check
+A signed integer (blessed or otherwise) that can safely hold its own number.  This is different than C<<< IntLike >>>, which doesn't check
 for storage limitations.
 
-=item C<< SignedInt[`b] >>
+=head3 SignedInt[`b]
 
-A signed integer that can hold a C<`b> bit number and is within those boundaries.  One bit is reserved for the sign, so the max limit
-on a 32-bit integer is actually C<< 2**31-1 >> or C<2147483647>.
+A signed integer that can hold a C<<< `b >>> bit number and is within those boundaries.  One bit is reserved for the sign, so the max limit
+on a 32-bit integer is actually C<<< 2**31-1 >>> or C<<< 2147483647 >>>.
 
-=item C<< UnsignedInt >>
+=head3 UnsignedInt
 
-Like C<SignedInt>, but with a minimum boundary of zero.
+Like L</SignedInt>, but with a minimum boundary of zero.
 
-=item C<< UnsignedInt[`b] >>
+=head3 UnsignedInt[`b]
 
-Like C<< SignedInt[`b] >>, but for unsigned integers.  Also, unsigned integers gain their extra bit, so the maximum is twice as high.
-
-=back
+Like L</SignedInt[`b]>, but for unsigned integers.  Also, unsigned integers gain their extra bit, so the maximum is twice as high.
 
 =head2 Floating-point numbers
 
-=over
+=head3 PerlSafeFloat
 
-=item C<< PerlSafeFloat >>
-
-A Perl native float that is in the "integer safe" range, or is a NaN/Inf value.
+A Perl native float that is in the "integer safe" range, or is a NaNE<sol>Inf value.
 
 This doesn't guarantee that every single fractional number is going to retain all of its information here.  It only guarantees that
 the whole number will be retained, even if the fractional part is partly or completely lost.
 
-=item C<< BlessedFloat >>
+=head3 BlessedFloat
 
 A blessed number that will support fractional numbers.  A L<Math::BigFloat> number will pass, whereas a L<Math::BigInt> number will
 fail.  However, if that L<Math::BigInt> number is capable of upgrading to a L<Math::BigFloat>, it will pass.
 
-=item C<< BlessedFloat[`d] >>
+=head3 BlessedFloat[`d]
 
 A float-capable blessed number that supports at least certain amount of digit accuracy.  The number itself is not boundary checked, as
 it is excessively difficult to figure out the exact dimensions of a floating point number.  It would also not be useful for numbers
-like C<0.333333...> to fail checks.
+like C<<< 0.333333... >>> to fail checks.
 
-=item C<< FloatSafeNum >>
+=head3 FloatSafeNum
 
-A Union of C<PerlSafeFloat> and C<BlessedFloat>.  In other words, a float-capable number with some basic checks to make sure
+A Union of L</PerlSafeFloat> and L</BlessedFloat>.  In other words, a float-capable number with some basic checks to make sure
 information is retained.
 
-=item C<< FloatBinary[`b, `e] >>
+=head3 FloatBinary[`b, `e]
 
-A floating-point number that can hold a C<`b> bit number with C<`e> bits of exponent, and is within those boundaries (or is NaN/Inf).
+A floating-point number that can hold a C<<< `b >>> bit number with C<<< `e >>> bits of exponent, and is within those boundaries (or is NaNE<sol>Inf).
 The bit breakdown follows traditional IEEE 754 floating point standards.  For example:
 
-   FloatBinary[32, 8] =
-      32 bits total (`b)
-      23 bit  mantissa (significand precision)
-       8 bit  exponent (`e)
-       1 bit  sign (+/-)
+    FloatBinary[32, 8] =
+       32 bits total (`b)
+       23 bit  mantissa (significand precision)
+        8 bit  exponent (`e)
+        1 bit  sign (+/-)
 
-Unlike the C<*Int> types, if Perl's native number cannot support all dimensions of the floating-point number without losing
+Unlike the C<<< *Int >>> types, if Perl's native number cannot support all dimensions of the floating-point number without losing
 information, then unblessed numbers are completely off the table.  For example, assuming a 32-bit machine:
 
-   UnsignedInt[64]->check( 0 )        # pass
-   UnsignedInt[64]->check( 2 ** 30 )  # pass
-   UnsignedInt[64]->check( 2 ** 60 )  # fail, because 32-bit NVs can't safely hold it
+    UnsignedInt[64]->check( 0 )        # pass
+    UnsignedInt[64]->check( 2 ** 30 )  # pass
+    UnsignedInt[64]->check( 2 ** 60 )  # fail, because 32-bit NVs can't safely hold it
+ 
+    FloatBinary[64, 11]->check( 0 )    # fail
+    FloatBinary[64, 11]->check( $any_unblessed_number )  # fail
 
-   FloatBinary[64, 11]->check( 0 )    # fail
-   FloatBinary[64, 11]->check( $any_unblessed_number )  # fail
+=head3 FloatDecimal[`d, `e]
 
-=item C<< FloatDecimal[`d, `e] >>
-
-A floating-point number that can hold a C<`d> digit number with C<`e> digits of exponent.  Modeled after the IEEE 754 "decimal" float.
-Rejects all Perl NVs that won't support the dimensions.  (See C<< FloatBinary[`b, `e] >>.)
-
-=back
+A floating-point number that can hold a C<<< `d >>> digit number with C<<< `e >>> digits of exponent.  Modeled after the IEEE 754 "decimal" float.
+Rejects all Perl NVs that won't support the dimensions.  (See L</FloatBinary[`b, `e]>.)
 
 =head2 Fixed-point numbers
 
-=over
+=head3 RealSafeNum
 
-=item C<< RealSafeNum >>
+Like L</FloatSafeNum>, but rejects any NaNE<sol>Inf.
 
-Like C<FloatSafeNum>, but rejects any NaN/Inf.
+=head3 FixedBinary[`b, `s]
 
-=item C<< FixedBinary[`b, `s] >>
-
-A fixed-point number, represented as a C<`b> bit integer than has been shifted by C<`s> digits.  For example, a
-C<< FixedBinary[32, 4] >> has a max of C<< 2**31-1 / 10**4 = 214748.3647 >>.  Because integers do not hold NaN/Inf, this type fails
+A fixed-point number, represented as a C<<< `b >>> bit integer than has been shifted by C<<< `s >>> digits.  For example, a
+C<<< FixedBinary[32, 4] >>> has a max of C<<< 2**31-1 / 10**4 = 214748.3647 >>>.  Because integers do not hold NaNE<sol>Inf, this type fails
 on those.
 
-Otherwise, it has the same properties and caveats as the parameterized C<< Float* >> types.
+Otherwise, it has the same properties and caveats as the parameterized C<<< Float* >>> types.
 
-=item C<< FixedDecimal[`d, `s] >>
+=head3 FixedDecimal[`d, `s]
 
-Like C<< FixedBinary[`b, `s] >>, but for a C<`d> digit integer.  Or, you could think of C<`d> and C<`s> as accuracy (significant
+Like L</FixedBinary[`b, `s]>, but for a C<<< `d >>> digit integer.  Or, you could think of C<<< `d >>> and C<<< `s >>> as accuracy (significant
 figures) and decimal precision, respectively.
 
-=back
-
 =head2 Characters
 
 Characters are basically encoded numbers, so there's a few types here.  If you need types that handle multi-length strings, you're
-better off using L<Types::Encoding>.
+better off using L<Types::Encodings>.
 
-=over
-
-=item C<< Char >>
+=head3 Char
 
 A single character.  Unicode is supported, but it must be decoded first.  A multi-byte character that Perl thinks is two separate
 characters will fail this type.
 
-=item C<< Char[`b] >>
+=head3 Char[`b]
 
-A single character that fits within C<`b> bits.  Unicode is supported, but it must be decoded first.
-
-Also, be aware of the ambiguous nature of 8-bit ASCII characters vs. UTF8:
-
-   use Encode qw(encode decode);
-
-   my $char = 'ђ';
-   Char[8]->check($char);   # pass
-   print ord($char);        # 209
-
-   $char = decode("UTF-8", $char);
-   Char[8]->check($char);   # fail
-   print ord($char);        # 1106
-   print $char;             # ђ
-
-To mitigate this effect, consider an intersection with C<Chars> and possibly a C<Decode> coercion (both from L<Types::Encoding>).
-
-=back
+A single character that fits within C<<< `b >>> bits.  Unicode is supported, but it must be decoded first.
 
 =head1 AVAILABILITY
 
